# Copyright (C) 2015 Twitter, Inc.

"""Container for all helpers and utilities used throughout the Ads API SDK."""

from datetime import timedelta
from email.utils import formatdate
from time import mktime

from twitter_ads import VERSION
from twitter_ads.enum import GRANULARITY


def get_version():
    """Returns a string representation of the current SDK version."""
    if isinstance(VERSION[-1], str):
        return '.'.join(map(str, VERSION[:-1])) + VERSION[-1]
    return '.'.join(map(str, VERSION))


def to_time(time, granularity):
    """Returns a truncated and rounded time string based on the specified granularity."""
    if not granularity:
        return format_time(time)
    if granularity == GRANULARITY.HOUR:
        return format_time(time - timedelta(
            minutes=time.minute, seconds=time.second, microseconds=time.microsecond))
    elif granularity == GRANULARITY.DAY:
        return format_time(time - timedelta(
            hours=time.hour, minutes=time.minute,
            seconds=time.second, microseconds=time.microsecond))
    else:
        return format_time(time)


def format_time(time):
    """Formats a datetime as an ISO 8601 compliant string."""
    return time.strftime('%Y-%m-%dT%H:%M:%SZ')


def http_time(time):
    """Formats a datetime as an RFC 1123 compliant string."""
    return formatdate(timeval=mktime(time.timetuple()), localtime=False, usegmt=True)



# Copyright (C) 2015 Twitter, Inc.

"""Container for all plugable resource object logic used by the Ads API SDK."""

import dateutil.parser
from datetime import datetime, timedelta

from twitter_ads.utils import format_time, to_time
from twitter_ads.enum import TRANSFORM, GRANULARITY
from twitter_ads.http import Request
from twitter_ads.cursor import Cursor


def resource_property(klass, name, **kwargs):
        """Builds a resource object property."""
        klass.PROPERTIES[name] = kwargs

        def getter(self):
            return getattr(self, '_%s' % name, kwargs.get('default', None))

        if kwargs.get('readonly', False):
            setattr(klass, name, property(getter))
        else:
            def setter(self, value):
                setattr(self, '_%s' % name, value)
            setattr(klass, name, property(getter, setter))


class Resource(object):
    """Base class for all API resource objects."""

    def __init__(self, account):
        self._account = account

    @property
    def account(self):
        return self._account

    def from_response(self, response):
        """
        Populates a given objects attributes from a parsed JSON API response.
        This helper handles all necessary type coercions as it assigns
        attribute values.
        """
        for name in self.PROPERTIES:
            attr = '_{0}'.format(name)
            transform = self.PROPERTIES[name].get('transform', None)
            value = response.get(name, None)
            if transform and transform == TRANSFORM.TIME and value:
                setattr(self, attr, dateutil.parser.parse(value))
            if isinstance(value, int) and value == 0:
                continue  # skip attribute
            else:
                setattr(self, attr, value)

        return self

    def to_params(self):
        """
        Generates a Hash of property values for the current object. This helper
        handles all necessary type coercions as it generates its output.
        """
        params = {}
        for name in self.PROPERTIES:
            attr = '_{0}'.format(name)
            value = getattr(self, attr, None) or getattr(self, name, None)

            # skip attribute
            if value is None:
                continue

            if isinstance(value, datetime):
                params[name] = format_time(value)
            elif isinstance(value, list):
                params[name] = ','.join(map(str, value))
            elif isinstance(value, bool):
                params[name] = str(value).lower()
            else:
                params[name] = value

        return params

    @classmethod
    def all(klass, account, **kwargs):
        """Returns a Cursor instance for a given resource."""
        resource = klass.RESOURCE_COLLECTION.format(account_id=account.id)
        request = Request(account.client, 'get', resource, params=kwargs)
        return Cursor(klass, request, init_with=[account])

    @classmethod
    def load(klass, account, id, **kwargs):
        """Returns an object instance for a given resource."""
        resource = klass.RESOURCE.format(account_id=account.id, id=id)
        response = Request(account.client, 'get', resource, params=kwargs).perform()

        return klass(account).from_response(response.body['data'])

    def reload(self, **kwargs):
        """
        Reloads all attributes for the current object instance from the API.
        """
        if not self.id:
            return self

        resource = self.RESOURCE.format(account_id=self.account.id, id=self.id)
        response = Request(self.account.client, 'get', resource, params=kwargs).perform()

        return self.from_response(response.body['data'])

    def __repr__(self):
        return '<{name} resource at {mem} id={id}>'.format(
            name=self.__class__.__name__,
            mem=hex(id(self)),
            id=getattr(self, 'id', None)
        )

    def _validate_loaded(self):
        if not self.id:
            raise ValueError("""
            Error! {klass} object not yet initialized,
            call {klass}.load first.
            """).format(klass=self.__class__)

    def _load_resource(self, klass, id, **kwargs):
        self._validate_loaded()
        if id is None:
            return klass.all(self, **kwargs)
        else:
            return klass.load(self, id, **kwargs)


class Persistence(object):
    """
    Container for all persistence related logic used by API resource objects.
    """

    def save(self):
        """
        Saves or updates the current object instance depending on the
        presence of `object.id`.
        """
        if self.id:
            method = 'put'
            resource = self.RESOURCE.format(account_id=self.account.id, id=self.id)
        else:
            method = 'post'
            resource = self.RESOURCE_COLLECTION.format(account_id=self.account.id)

        response = Request(
            self.account.client, method,
            resource, params=self.to_params()).perform()

        return self.from_response(response.body['data'])

    def delete(self):
        """
        Deletes the current object instance depending on the
        presence of `object.id`.
        """
        resource = self.RESOURCE.format(account_id=self.account.id, id=self.id)
        response = Request(self.account.client, 'delete', resource).perform()
        self.from_response(response.body['data'])


class Analytics(object):
    """
    Container for all analytics related logic used by API resource objects.
    """

    ANALYTICS_MAP = {
        'LineItem': 'line_item_ids',
        'OrganicTweet': 'tweet_ids',
        'Tweet': 'tweet_ids',
        'PromotedTweet': 'promoted_tweet_ids'
    }

    def stats(self, metrics, **kwargs):  # noqa
        """
        Pulls a list of metrics for the current object instance.
        """
        return self.__class__.all_stats(self.account, [self.id], metrics, **kwargs)

    @classmethod
    def all_stats(klass, account, ids, metrics, **kwargs):
        """
        Pulls a list of metrics for a specified set of object IDs.
        """
        end_time = kwargs.get('end_time', datetime.utcnow())
        start_time = kwargs.get('start_time', end_time - timedelta(seconds=604800))
        granularity = kwargs.get('granularity', GRANULARITY.HOUR)
        segmentation_type = kwargs.get('segmentation_type', None)

        params = {
            'metrics': ','.join(metrics),
            'start_time': to_time(start_time, granularity),
            'end_time': to_time(end_time, granularity),
            'granularity': granularity.upper()
        }
        if segmentation_type is not None:
            params['segmentation_type'] = segmentation_type.upper()

        params[klass.ANALYTICS_MAP[klass.__name__]] = ','.join(ids)

        resource = klass.RESOURCE_STATS.format(account_id=account.id)
        response = Request(account.client, 'get', resource, params=params).perform()
        return response.body['data']



# Copyright (C) 2015 Twitter, Inc.

"""Container for all creative management logic used by the Ads API SDK."""

from twitter_ads.enum import TRANSFORM
from twitter_ads.resource import resource_property, Resource, Persistence, Analytics
from twitter_ads.http import Request


class PromotedAccount(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/promoted_accounts'
    RESOURCE_STATS = '/0/stats/accounts/{account_id}/promoted_accounts'
    RESOURCE = '/0/accounts/{account_id}/promoted_accounts/{id}'

# promoted account properties
# read-only
resource_property(PromotedAccount, 'id', readonly=True)
resource_property(PromotedAccount, 'approval_status', readonly=True)
resource_property(PromotedAccount, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(PromotedAccount, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(PromotedAccount, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(PromotedAccount, 'line_item_id')
resource_property(PromotedAccount, 'user_id')
resource_property(PromotedAccount, 'paused', transform=TRANSFORM.BOOL)


class PromotedTweet(Resource, Persistence, Analytics):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/promoted_tweets'
    RESOURCE_STATS = '/0/stats/accounts/{account_id}/promoted_tweets'
    RESOURCE = '/0/accounts/{account_id}/promoted_tweets/{id}'

    def save(self):
        """
        Saves or updates the current object instance depending on the
        presence of `object.id`.
        """
        params = self.to_params()
        if 'tweet_id' in params:
            params['tweet_ids'] = [params['tweet_id']]
            del params['tweet_id']

        if self.id:
            resource = self.RESOURCE.format(account_id=self.account.id, id=self.id)
            response = Request(self.account.client, 'put', resource, params=params).perform()
            return self.from_response(response.body['data'])

        resource = self.RESOURCE_COLLECTION.format(account_id=self.account.id)
        response = Request(self.account.client, 'post', resource, params=params).perform()
        return self.from_response(response.body['data'][0])

# promoted tweet properties
# read-only
resource_property(PromotedTweet, 'id', readonly=True)
resource_property(PromotedTweet, 'approval_status', readonly=True)
resource_property(PromotedTweet, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(PromotedTweet, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(PromotedTweet, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(PromotedTweet, 'line_item_id')
resource_property(PromotedTweet, 'tweet_id')
resource_property(PromotedTweet, 'paused', transform=TRANSFORM.BOOL)


class Video(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/videos'
    RESOURCE = '/0/accounts/{account_id}/videos/{id}'

# video properties
# read-only
resource_property(Video, 'id', readonly=True)
resource_property(Video, 'tweeted', readonly=True, transform=TRANSFORM.BOOL)
resource_property(Video, 'ready_to_tweet', readonly=True, transform=TRANSFORM.BOOL)
resource_property(Video, 'duration', readonly=True)
resource_property(Video, 'reasons_not_servable', readonly=True)
resource_property(Video, 'preview_url', readonly=True)
resource_property(Video, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(Video, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(Video, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(Video, 'title')
resource_property(Video, 'description')
resource_property(Video, 'video_media_id')


class WebsiteCard(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/cards/website'
    RESOURCE = '/0/accounts/{account_id}/cards/website/{id}'

# website card properties
# read-only
resource_property(WebsiteCard, 'id', readonly=True)
resource_property(WebsiteCard, 'preview_url', readonly=True)
resource_property(WebsiteCard, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(WebsiteCard, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(WebsiteCard, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(WebsiteCard, 'name')
resource_property(WebsiteCard, 'website_title')
resource_property(WebsiteCard, 'website_url')
resource_property(WebsiteCard, 'website_cta')
resource_property(WebsiteCard, 'image_media_id')


class LeadGenCard(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/cards/lead_gen'
    RESOURCE = '/0/accounts/{account_id}/cards/lead_gen/{id}'

# lead gen card properties
# read-only
resource_property(LeadGenCard, 'id', readonly=True)
resource_property(LeadGenCard, 'preview_url', readonly=True)
resource_property(LeadGenCard, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(LeadGenCard, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(LeadGenCard, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(LeadGenCard, 'name')
resource_property(LeadGenCard, 'image_media_id')
resource_property(LeadGenCard, 'cta')
resource_property(LeadGenCard, 'fallback_url')
resource_property(LeadGenCard, 'privacy_policy_url')
resource_property(LeadGenCard, 'title')
resource_property(LeadGenCard, 'submit_url')
resource_property(LeadGenCard, 'submit_method')
resource_property(LeadGenCard, 'custom_destination_url')
resource_property(LeadGenCard, 'custom_destination_text')
resource_property(LeadGenCard, 'custom_key_screen_name')
resource_property(LeadGenCard, 'custom_key_name')
resource_property(LeadGenCard, 'custom_key_email')


class AppDownloadCard(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/cards/app_download'
    RESOURCE = '/0/accounts/{account_id}/cards/app_download/{id}'

# app download card properties
# read-only
resource_property(AppDownloadCard, 'id', readonly=True)
resource_property(AppDownloadCard, 'preview_url', readonly=True)
resource_property(AppDownloadCard, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(AppDownloadCard, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(AppDownloadCard, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(AppDownloadCard, 'name')
resource_property(AppDownloadCard, 'app_country_code')
resource_property(AppDownloadCard, 'iphone_app_id')
resource_property(AppDownloadCard, 'iphone_deep_link')
resource_property(AppDownloadCard, 'ipad_app_id')
resource_property(AppDownloadCard, 'ipad_deep_link')
resource_property(AppDownloadCard, 'googleplay_app_id')
resource_property(AppDownloadCard, 'googleplay_deep_link')
resource_property(AppDownloadCard, 'app_cta')
resource_property(AppDownloadCard, 'custom_icon_media_id')
resource_property(AppDownloadCard, 'custom_app_description')


class ImageAppDownloadCard(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/cards/image_app_download'
    RESOURCE = '/0/accounts/{account_id}/cards/image_app_download/{id}'

# image app download card properties
# read-only
resource_property(ImageAppDownloadCard, 'id', readonly=True)
resource_property(ImageAppDownloadCard, 'preview_url', readonly=True)
resource_property(ImageAppDownloadCard, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(ImageAppDownloadCard, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(ImageAppDownloadCard, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(ImageAppDownloadCard, 'name')
resource_property(ImageAppDownloadCard, 'app_country_code')
resource_property(ImageAppDownloadCard, 'iphone_app_id')
resource_property(ImageAppDownloadCard, 'iphone_deep_link')
resource_property(ImageAppDownloadCard, 'ipad_app_id')
resource_property(ImageAppDownloadCard, 'ipad_deep_link')
resource_property(ImageAppDownloadCard, 'googleplay_app_id')
resource_property(ImageAppDownloadCard, 'googleplay_deep_link')
resource_property(ImageAppDownloadCard, 'app_cta')
resource_property(ImageAppDownloadCard, 'wide_app_image_media_id')


class VideoAppDownloadCard(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/cards/video_app_download'
    RESOURCE = '/0/accounts/{account_id}/cards/video_app_download/{id}'

# video app download card properties
# read-only
resource_property(VideoAppDownloadCard, 'id', readonly=True)
resource_property(VideoAppDownloadCard, 'preview_url', readonly=True)
resource_property(VideoAppDownloadCard, 'video_url', readonly=True)
resource_property(VideoAppDownloadCard, 'video_poster_url', readonly=True)
resource_property(VideoAppDownloadCard, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(VideoAppDownloadCard, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(VideoAppDownloadCard, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(VideoAppDownloadCard, 'name')
resource_property(VideoAppDownloadCard, 'app_country_code')
resource_property(VideoAppDownloadCard, 'iphone_app_id')
resource_property(VideoAppDownloadCard, 'iphone_deep_link')
resource_property(VideoAppDownloadCard, 'ipad_app_id')
resource_property(VideoAppDownloadCard, 'ipad_deep_link')
resource_property(VideoAppDownloadCard, 'googleplay_app_id')
resource_property(VideoAppDownloadCard, 'googleplay_deep_link')
resource_property(VideoAppDownloadCard, 'app_cta')
resource_property(VideoAppDownloadCard, 'image_media_id')
resource_property(VideoAppDownloadCard, 'video_id')


class ImageConversationCard(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/cards/image_conversation'
    RESOURCE = '/0/accounts/{account_id}/cards/image_conversation/{id}'

# image conversation card properties
# read-only
resource_property(ImageConversationCard, 'id', readonly=True)
resource_property(ImageConversationCard, 'preview_url', readonly=True)
resource_property(ImageConversationCard, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(ImageConversationCard, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(ImageConversationCard, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(ImageConversationCard, 'name')
resource_property(ImageConversationCard, 'title')
resource_property(ImageConversationCard, 'first_cta')
resource_property(ImageConversationCard, 'first_cta_tweet')
resource_property(ImageConversationCard, 'second_cta')
resource_property(ImageConversationCard, 'second_cta_tweet')
resource_property(ImageConversationCard, 'thank_you_text')
resource_property(ImageConversationCard, 'thank_you_url')
resource_property(ImageConversationCard, 'image_media_id')


class VideoConversationCard(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/cards/video_conversation'
    RESOURCE = '/0/accounts/{account_id}/cards/video_conversation/{id}'

# video conversation card properties
# read-only
resource_property(VideoConversationCard, 'id', readonly=True)
resource_property(VideoConversationCard, 'preview_url', readonly=True)
resource_property(VideoConversationCard, 'video_url', readonly=True)
resource_property(VideoConversationCard, 'video_poster_url', readonly=True)
resource_property(VideoConversationCard, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(VideoConversationCard, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(VideoConversationCard, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(VideoConversationCard, 'name')
resource_property(VideoConversationCard, 'title')
resource_property(VideoConversationCard, 'first_cta')
resource_property(VideoConversationCard, 'first_cta_tweet')
resource_property(VideoConversationCard, 'second_cta')
resource_property(VideoConversationCard, 'second_cta_tweet')
resource_property(VideoConversationCard, 'thank_you_text')
resource_property(VideoConversationCard, 'thank_you_url')
resource_property(VideoConversationCard, 'image_media_id')
resource_property(VideoConversationCard, 'video_id')



# Copyright (C) 2015 Twitter, Inc.

"""Container for all audience management logic used by the Ads API SDK."""

from twitter_ads.enum import TA_OPERATIONS, TRANSFORM
from twitter_ads.resource import resource_property, Resource
from twitter_ads.http import TONUpload, Request
from twitter_ads.error import BadRequest
from twitter_ads.cursor import Cursor


class TailoredAudience(Resource):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/tailored_audiences'
    RESOURCE = '/0/accounts/{account_id}/tailored_audiences/{id}'
    RESOURCE_UPDATE = '/0/accounts/{account_id}/tailored_audience_changes'
    OPT_OUT = '/0/accounts/{account_id}/tailored_audiences/global_opt_out'

    @classmethod
    def create(klass, account, file_path, name, list_type):
        """
        Uploads and creates a new tailored audience.
        """
        upload = TONUpload(account.client, file_path)
        audience = klass(account)
        getattr(audience, '__create_audience__')(name, list_type)
        try:
            getattr(audience, '__update_audience__')(upload.perform(), list_type, TA_OPERATIONS.ADD)
            return audience.reload()
        except BadRequest as e:
            audience.delete()
            raise e

    @classmethod
    def opt_out(klass, account, file_path, list_type):
        """
        Updates the global opt-out list for the specified advertiser account.
        """
        upload = TONUpload(account.client, file_path)
        params = {'input_file_path': upload.perform(), 'list_type': list_type}
        resource = klass.OPT_OUT.format(account_id=account.id)
        Request(account.client, 'put', resource, params=params).perform()
        return True

    def update(self, file_path, list_type, operation=TA_OPERATIONS.ADD):
        """
        Updates the current tailored audience instance.
        """
        upload = TONUpload(self.account.client, file_path)
        getattr(self, '__update_audience__')(upload.perform(), list_type, operation)
        return self.reload()

    def delete(self):
        """
        Deletes the current tailored audience instance.
        """
        resource = self.RESOURCE.format(account_id=self.account.id, id=self.id)
        response = Request(self.account.client, 'delete', resource).perform()
        return self.from_response(response.body['data'])

    def status(self):
        """
        Returns the status of all changes for the current tailored audience instance.
        """
        if not self.id:
            return None

        resource = self.RESOURCE_UPDATE.format(account_id=self.account.id)
        request = Request(self.account.client, 'get', resource, params=self.to_params())
        cursor = list(Cursor(None, request))

        return filter(lambda change: change['tailored_audience_id'] == self.id, cursor)

    def __create_audience__(self, name, list_type):
        params = {'name': name, 'list_type': list_type}
        resource = self.RESOURCE_COLLECTION.format(account_id=self.account.id)
        response = Request(self.account.client, 'post', resource, params=params).perform()
        return self.from_response(response.body['data'])

    def __update_audience__(self, location, list_type, operation):
        params = {
            'tailored_audience_id': self.id,
            'input_file_path': location,
            'list_type': list_type,
            'operation': operation
        }

        resource = self.RESOURCE_UPDATE.format(account_id=self.account.id)
        return Request(self.account.client, 'post', resource, params=params).perform()

# tailored audience properties
# read-only
resource_property(TailoredAudience, 'id', readonly=True)
resource_property(TailoredAudience, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(TailoredAudience, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(TailoredAudience, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
resource_property(TailoredAudience, 'audience_size', readonly=True)
resource_property(TailoredAudience, 'audience_type', readonly=True)
resource_property(TailoredAudience, 'metadata', readonly=True)
resource_property(TailoredAudience, 'partner_source', readonly=True)
resource_property(TailoredAudience, 'reasons_not_targetable', readonly=True)
resource_property(TailoredAudience, 'targetable', readonly=True)
resource_property(TailoredAudience, 'targetable_types', readonly=True)
# writable
resource_property(TailoredAudience, 'name')
resource_property(TailoredAudience, 'list_type')



# Copyright (C) 2015 Twitter, Inc.

"""Container for all HTTP request and response logic for the SDK."""

import os
import sys
import platform
import logging
import json

try:
    import httplib2 as httplib
except ImportError:
    if sys.version_info[0] != 3:
        import httplib
    else:
        import http.client as httplib

import dateutil.parser
from datetime import datetime, timedelta

from requests_oauthlib import OAuth1Session

from twitter_ads.utils import get_version, http_time
from twitter_ads.error import Error


class Request(object):
    """Generic container for all API requests."""

    _DEFAULT_DOMAIN = 'https://ads-api.twitter.com'
    _SANDBOX_DOMAIN = 'https://ads-api-sandbox.twitter.com'

    _HTTP_METHOD = [
        'get',
        'put',
        'post',
        'delete',
    ]

    def __init__(self, client, method, resource, **kwargs):
        self._client = client
        self._resource = resource
        self._options = kwargs.copy()

        method = method.lower()
        if method not in self._HTTP_METHOD:
            msg = "Error! {0} is not an allowed HTTP method type.".format(method)
            raise ValueError(msg)

        self._method = method

    @property
    def options(self):
        return self._options

    @property
    def client(self):
        return self._client

    @property
    def method(self):
        return self._method

    @property
    def resource(self):
        return self._resource

    def perform(self):
        if self.client.trace:
            self.__enable_logging()
        response = self.__oauth_request()
        if response.code > 399:
            raise Error.from_response(response)
        return response

    def __oauth_request(self):
        headers = {'user-agent': self.__user_agent()}
        if 'headers' in self.options:
            headers.update(self.options['headers'].copy())

        params = self.options.get('params', None)
        data = self.options.get('body', None)
        files = self.options.get('files', None)

        consumer = OAuth1Session(
            self._client.consumer_key,
            client_secret=self._client.consumer_secret,
            resource_owner_key=self._client.access_token,
            resource_owner_secret=self._client.access_token_secret)

        url = self.__domain() + self._resource
        method = getattr(consumer, self._method)
        response = method(url, headers=headers, data=data, params=params, files=files)

        return Response(response.status_code, response.headers,
                        body=response.raw, raw_body=response.text)

    def __enable_logging(self):
        httplib.HTTPConnection.debuglevel = 1
        logging.basicConfig(level=logging.DEBUG)
        logging.propagate = True

    def __user_agent(self):
        python_verison = "{0}.{1}".format(sys.version_info.major, sys.version_info.minor)
        return 'twitter-ads version: {0} platform: Python {1} ({2}/{3})'.format(
            get_version(),
            python_verison,
            platform.python_implementation(),
            sys.platform)

    def __domain(self):
        if self.client.sandbox:
            return self._SANDBOX_DOMAIN
        return self.options.get('domain', self._DEFAULT_DOMAIN)


class Response(object):
    """Generic container for API responses."""

    def __init__(self, code, headers, **kwargs):
        self._code = code
        self._headers = headers
        self._raw_body = kwargs.get('raw_body', None)

        try:
            self._body = json.loads(self._raw_body)
        except ValueError:
            self._body = self._raw_body

        if 'x-rate-limit-reset' in headers:
            self._rate_limit = int(headers['x-rate-limit-limit'])
            self._rate_limit_remaining = int(headers['x-rate-limit-remaining'])
            self._rate_limit_reset = datetime.fromtimestamp(int(headers['x-rate-limit-reset']))
        elif 'x-cost-rate-limit-reset' in headers:
            self._rate_limit = int(headers['x-cost-rate-limit-limit'])
            self._rate_limit_remaining = int(headers['x-cost-rate-limit-remaining'])
            self._rate_limit_reset = dateutil.parser.parse(headers['x-cost-rate-limit-reset'].first)
        else:
            self._rate_limit = None
            self._rate_limit_remaining = None
            self._rate_limit_reset = None

    @property
    def code(self):
        return self._code

    @property
    def headers(self):
        return self._headers

    @property
    def body(self):
        return self._body

    @property
    def raw_body(self):
        return self._raw_body

    @property
    def rate_limit(self):
        return self._rate_limit

    @property
    def rate_limit_remaining(self):
        return self._rate_limit_remaining

    @property
    def rate_limit_reset(self):
        return self._rate_limit_reset

    @property
    def error(self):
        return True if (self._code >= 400 and self._code <= 599) else False


class TONUpload(object):
    """Specialized request class for TON API uploads."""

    _DEFAULT_DOMAIN = 'https://ton.twitter.com'
    _DEFAULT_RESOURCE = '/1.1/ton/bucket/'
    _DEFAULT_BUCKET = 'ta_partner'
    _DEFAULT_EXPIRE = datetime.now() + timedelta(days=10)
    _MIN_FILE_SIZE = 1024 * 1024 * 1

    def __init__(self, client, file_path, **kwargs):
        if not os.path.isfile(file_path):
            msg = "Error! The specified file does not exist. ({0})".format(file_path)
            raise ValueError(msg)

        self._file_path = os.path.abspath(file_path)
        self._file_size = os.path.getsize(self._file_path)
        self._client = client
        self._options = kwargs.copy()
        self._bucket = self._options.pop('bucket', self._DEFAULT_BUCKET)

    @property
    def options(self):
        return self._options

    @property
    def client(self):
        return self._client

    @property
    def bucket(self):
        return self._bucket

    @property
    def content_type(self):
        """Returns the content-type value determined by file extension."""

        if hasattr(self, '_content_type'):
            return self._content_type

        filename, extension = os.path.splitext(self._file_path)
        if extension == '.csv':
            self._content_type = 'text/csv'
        elif extension == '.tsv':
            self._content_type = 'text/tab-separated-values'
        else:
            self._content_type = 'text/plain'

        return self._content_type

    def perform(self):
        """Executes the current TONUpload object."""

        if self._file_size < self._MIN_FILE_SIZE:
            resource = "{0}{1}".format(self._DEFAULT_RESOURCE, self.bucket)
            response = self.__upload(resource, open(self._file_path, 'rb').read())
            return response.headers['location']
        else:
            response = self.__init_chunked_upload()
            chunk_size = int(response.headers['x-ton-min-chunk-size'])
            location = response.headers['location']

            f = open(self._file_path, 'rb')
            bytes_read = 0
            while True:
                bytes = f.read(chunk_size)
                if not bytes:
                    break
                bytes_start = bytes_read
                bytes_read += len(bytes)
                self.__upload_chunk(location, chunk_size, bytes, bytes_start, bytes_read)
            f.close()

            return location

    def __repr__(self):
        return '<{name} object at {mem} bucket={bucket} file={file}>'.format(
            name=self.__class__.__name__,
            mem=hex(id(self)),
            bucket=self.bucket,
            file=self._file_path
        )

    def __upload(self, resource, bytes):
        """Performs a single chunk upload."""

        # note: string conversion required here due to open encoding bug in requests-oauthlib.
        headers = {
            'x-ton-expires': http_time(self.options.get('x-ton-expires', self._DEFAULT_EXPIRE)),
            'content-length': str(self._file_size),
            'content-type': self.content_type
        }

        return Request(self._client, 'post', resource,
                       domain=self._DEFAULT_DOMAIN, headers=headers, body=bytes).perform()

    def __init_chunked_upload(self):
        """Initialization for a multi-chunk upload."""

        # note: string conversion required here due to open encoding bug in requests-oauthlib.
        headers = {
            'x-ton-content-type': self.content_type,
            'x-ton-content-length': str(self._file_size),
            'x-ton-expires': http_time(self.options.get('x-ton-expires', self._DEFAULT_EXPIRE)),
            'content-length': str(0),
            'content-type': self.content_type
        }

        resource = "{0}{1}?resumable=true".format(self._DEFAULT_RESOURCE, self._DEFAULT_BUCKET)

        return Request(self._client, 'post', resource,
                       domain=self._DEFAULT_DOMAIN, headers=headers).perform()

    def __upload_chunk(self, resource, chunk_size, bytes, bytes_start, bytes_read):
        """Uploads a single chunk of a multi-chunk upload."""

        # note: string conversion required here due to open encoding bug in requests-oauthlib.
        headers = {
            'content-type': self.content_type,
            'content-length': str(min([chunk_size, self._file_size - bytes_read])),
            'content-range': "bytes {0}-{1}/{2}".format(
                bytes_start, bytes_read - 1, self._file_size)
        }

        return Request(self._client, 'put', resource,
                       domain=self._DEFAULT_DOMAIN, headers=headers, body=bytes).perform()



# Copyright (C) 2015 Twitter, Inc.

"""Container for all enum values used by the Ads API SDK."""


def enum(**enums):
    return type('Enum', (), enums)

TRANSFORM = enum(
    TIME=0,
    BOOL=1,
    INT=2,
    LIST=3,
    OBJECT=4
)

BID_UNIT = enum(
    APP_CLICK='APP_CLICK',
    APP_INSTALL='APP_INSTALL',
    ENGAGEMENT='ENGAGEMENT',
    FOLLOW='FOLLOW',
    LEAD='LEAD',
    LINK_CLICK='LINK_CLICK',
    VIEW='VIEW'
)

CHARGE_BY = enum(
    APP_CLICK='APP_CLICK',
    APP_INSTALL='APP_INSTALL',
    ENGAGEMENT='ENGAGEMENT',
    FOLLOW='FOLLOW',
    LEAD='LEAD',
    LINK_CLICK='LINK_CLICK',
    VIEW='VIEW'
)

OPTIMIZATIONS = enum(
    DEFAULT='DEFAULT',
    WEBSITE_CONVERSIONS='WEBSITE_CONVERSIONS'
)

PRODUCT = enum(
    PROMOTED_ACCOUNT='PROMOTED_ACCOUNT',
    PROMOTED_TWEETS='PROMOTED_TWEETS'
)

PLACEMENT = enum(
    ALL_ON_TWITTER='ALL_ON_TWITTER',
    TWITTER_SEARCH='TWITTER_SEARCH',
    TWITTER_TIMELINE='TWITTER_TIMELINE',
    PUBLISHER_NETWORK='PUBLISHER_NETWORK'
)

OBJECTIVE = enum(
    APP_ENGAGEMENTS='APP_ENGAGEMENTS',
    APP_INSTALLS='APP_INSTALLS',
    FOLLOWERS='FOLLOWERS',
    LEAD_GENERATION='LEAD_GENERATION',
    TWEET_ENGAGEMENTS='TWEET_ENGAGEMENTS',
    VIDEO_VIEWS='VIDEO_VIEWS',
    WEBSITE_CLICKS='WEBSITE_CLICKS',
    CUSTOM='CUSTOM'
)

GRANULARITY = enum(
    HOUR='HOUR',
    DAY='DAY',
    TOTAL='TOTAL'
)

AGE_BUCKET = enum(
    AGE_13_TO_17='AGE_13_TO_17',
    AGE_18_TO_24='AGE_18_TO_24',
    AGE_25_TO_34='AGE_25_TO_34',
    AGE_35_TO_44='AGE_35_TO_44',
    AGE_45_TO_54='AGE_45_TO_54',
    AGE_55_TO_64='AGE_55_TO_64',
    AGE_OVER_65='AGE_OVER_65'
)

AGE_BUCKET_COARSE = enum(
    AGE_18_TO_34='AGE_18_TO_34',
    AGE_18_TO_49='AGE_18_TO_49',
    AGE_25_TO_54='AGE_25_TO_54',
    AGE_OVER_21='AGE_OVER_21'
)

EVENTS = enum(
    MUSIC_AND_ENTERTAINMENT='MUSIC_AND_ENTERTAINMENT',
    SPORTS='SPORTS',
    HOLIDAY='HOLIDAY',
    CONFERENCE='CONFERENCE',
    OTHER='OTHER'
)

TA_LIST_TYPES = enum(
    EMAIL='EMAIL',
    DEVICE_ID='DEVICE_ID',
    TWITTER_ID='TWITTER_ID',
    HANDLE='HANDLE',
    PHONE_NUMBER='PHONE_NUMBER'
)

TA_OPERATIONS = enum(
    ADD='ADD',
    REMOVE='REMOVE',
    REPLACE='REPLACE'
)



# Copyright (C) 2015 Twitter, Inc.

"""Container for all campaign management logic used by the Ads API SDK."""

from twitter_ads.enum import TRANSFORM
from twitter_ads.resource import resource_property, Resource, Persistence, Analytics
from twitter_ads.http import Request
from twitter_ads.cursor import Cursor


class TargetingCriteria(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/targeting_criteria'
    RESOURCE = '/0/accounts/{account_id}/targeting_criteria/{id}'

    @classmethod
    def all(klass, account, line_item_id, **kwargs):
        """Returns a Cursor instance for a given resource."""
        params = {'line_item_id': line_item_id}
        params.update(kwargs)

        resource = klass.RESOURCE_COLLECTION.format(account_id=account.id)
        request = Request(account.client, 'get', resource, params=params)

        return Cursor(klass, request, init_with=[account])

# targeting criteria properties
# read-only
resource_property(TargetingCriteria, 'id', readonly=True)
resource_property(TargetingCriteria, 'localized_name', readonly=True)
resource_property(TargetingCriteria, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(TargetingCriteria, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(TargetingCriteria, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(TargetingCriteria, 'line_item_id')
resource_property(TargetingCriteria, 'targeting_type')
resource_property(TargetingCriteria, 'targeting_value')
resource_property(TargetingCriteria, 'tailored_audience_expansion')
resource_property(TargetingCriteria, 'tailored_audience_type')


class FundingInstrument(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/funding_instruments'
    RESOURCE = '/0/accounts/{account_id}/funding_instruments/{id}'

# funding instrument properties
# read-only
resource_property(FundingInstrument, 'id', readonly=True)
resource_property(FundingInstrument, 'name', readonly=True)
resource_property(FundingInstrument, 'cancelled', readonly=True, transform=TRANSFORM.BOOL)
resource_property(FundingInstrument, 'credit_limit_local_micro', readonly=True)
resource_property(FundingInstrument, 'currency', readonly=True)
resource_property(FundingInstrument, 'description', readonly=True)
resource_property(FundingInstrument, 'funded_amount_local_micro', readonly=True)
resource_property(FundingInstrument, 'type', readonly=True)
resource_property(FundingInstrument, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(FundingInstrument, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(FundingInstrument, 'deleted', readonly=True, transform=TRANSFORM.BOOL)


class PromotableUser(Resource):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/promotable_users'
    RESOURCE = '/0/accounts/{account_id}/promotable_users/{id}'

# promotable user properties
# read-only
resource_property(PromotableUser, 'id', readonly=True)
resource_property(PromotableUser, 'promotable_user_type', readonly=True)
resource_property(PromotableUser, 'user_id', readonly=True)
resource_property(PromotableUser, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(PromotableUser, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(PromotableUser, 'deleted', readonly=True, transform=TRANSFORM.BOOL)


class AppList(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/app_lists'
    RESOURCE = '/0/accounts/{account_id}/app_lists/{id}'

    def create(self, name, *ids):
        if isinstance(ids, list):
            ids = ','.join(map(str, ids))

        resource = self.RESOURCE_COLLECTION.format(account_id=self.account.id)
        params = self.to_params.update({'app_store_identifiers': ids, 'name': name})
        response = Request(self.account.client, 'post', resource, params=params).perform()

        return self.from_response(response.body['data'])

    def apps(self):
        if self.id and not hasattr(self, '_apps'):
            self.reload()
        return self._apps

# app list properties
# read-only
resource_property(AppList, 'id', readonly=True)
resource_property(AppList, 'name', readonly=True)
resource_property(AppList, 'apps', readonly=True)


class Campaign(Resource, Persistence):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/campaigns'
    RESOURCE_STATS = '/0/stats/accounts/{account_id}/campaigns'
    RESOURCE = '/0/accounts/{account_id}/campaigns/{id}'

# campaign properties
# read-only
resource_property(Campaign, 'id', readonly=True)
resource_property(Campaign, 'reasons_not_servable', readonly=True)
resource_property(Campaign, 'servable', readonly=True, transform=TRANSFORM.BOOL)
resource_property(Campaign, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(Campaign, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(Campaign, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(Campaign, 'name')
resource_property(Campaign, 'funding_instrument_id')
resource_property(Campaign, 'start_time', transform=TRANSFORM.TIME)
resource_property(Campaign, 'end_time', transform=TRANSFORM.TIME)
resource_property(Campaign, 'paused', transform=TRANSFORM.BOOL)
resource_property(Campaign, 'currency')
resource_property(Campaign, 'standard_delivery')
resource_property(Campaign, 'daily_budget_amount_local_micro')
resource_property(Campaign, 'total_budget_amount_local_micro')


class LineItem(Resource, Persistence, Analytics):

    PROPERTIES = {}

    RESOURCE_COLLECTION = '/0/accounts/{account_id}/line_items'
    RESOURCE_STATS = '/0/stats/accounts/{account_id}/line_items'
    RESOURCE = '/0/accounts/{account_id}/line_items/{id}'

    def targeting_criteria(self, id=None, **kwargs):
        """
        Returns a collection of targeting criteria available to the
        current line item.
        """
        self._validate_loaded()
        if id is None:
            return TargetingCriteria.all(self.account, self.id, **kwargs)
        else:
            return TargetingCriteria.load(self.account, id, **kwargs)

# line item properties
# read-only
resource_property(LineItem, 'id', readonly=True)
resource_property(LineItem, 'created_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(LineItem, 'updated_at', readonly=True, transform=TRANSFORM.TIME)
resource_property(LineItem, 'deleted', readonly=True, transform=TRANSFORM.BOOL)
# writable
resource_property(LineItem, 'name')
resource_property(LineItem, 'campaign_id')
resource_property(LineItem, 'advertiser_domain')
resource_property(LineItem, 'categories', transform=TRANSFORM.LIST)
resource_property(LineItem, 'charge_by')
resource_property(LineItem, 'include_sentiment')
resource_property(LineItem, 'objective')
resource_property(LineItem, 'optimization')
resource_property(LineItem, 'paused', transform=TRANSFORM.BOOL)
resource_property(LineItem, 'primary_web_event_tag')
resource_property(LineItem, 'product_type')
resource_property(LineItem, 'placements', transform=TRANSFORM.LIST)
resource_property(LineItem, 'bid_unit')
resource_property(LineItem, 'automatically_select_bid', transform=TRANSFORM.BOOL)
resource_property(LineItem, 'bid_amount_local_micro')
resource_property(LineItem, 'total_budget_amount_local_micro')


class Tweet(object):

    TWEET_PREVIEW = '/0/accounts/{account_id}/tweet/preview'
    TWEET_ID_PREVIEW = '/0/accounts/{account_id}/tweet/preview/{id}'
    TWEET_CREATE = '/0/accounts/{account_id}/tweet'

    def __init__(self):
        raise NotImplementedError(
            'Error! {name} cannot be instantiated.'.format(name=self.__class__.__name__))

    @classmethod
    def preview(klass, account, **kwargs):
        """
        Returns an HTML preview of a tweet, either new or existing.
        """
        params = {}
        params.update(kwargs)

        # handles array to string conversion for media IDs
        if 'media_ids' in params and isinstance(params['media_ids'], list):
            params['media_ids'] = ','.join(map(params['media_ids']))

        resource = klass.TWEET_ID_PREVIEW if params.get('id') else klass.TWEET_PREVIEW
        resource = resource.format(account_id=account.id, id=params.get('id'))
        response = Request(account.client, 'get', resource, params=params).perform()
        return response.body['data']

    @classmethod
    def create(klass, account, status, **kwargs):
        """
        Creates a "Promoted-Only" Tweet using the specialized Ads API end point.
        """
        params = {'status': status}
        params.update(kwargs)

        # handles array to string conversion for media IDs
        if 'media_ids' in params and isinstance(params['media_ids'], list):
            params['media_ids'] = ','.join(map(params['media_ids']))

        resource = klass.TWEET_CREATE.format(account_id=account.id)
        response = Request(account.client, 'post', resource, params=params).perform()
        return response.body['data']
